from django.core.management.base import BaseCommand

from django_bundles.conf.bundles_settings import bundles_settings
from django_bundles.core import get_bundles, set_bundle_versions
from django_bundles.processors import processor_pipeline, processor_library
from django_bundles.utils.files import FileChunkGenerator
from django_bundles.utils.processes import run_process

import os
import collections
from hashlib import md5
from tempfile import NamedTemporaryFile
from optparse import make_option


def iter_bundle_files(bundle):
    for bundle_file in bundle.files:
        for chunk in processor_pipeline(bundle_file.processors, FileChunkGenerator(open(bundle_file.file_path, 'rb'))):
            yield chunk
        yield '\n'


def make_uglify_bundle(bundle, fixed_version=None):
    m = md5()

    infile_list = []
    source_map_processed_input_files = []

    try:
        for bundle_file in bundle.files:
            if bundle_file.processors:
                # for now preprocessed files are written to temp files and therefore won't be available in the source map
                output_pipeline = processor_pipeline(bundle_file.processors, FileChunkGenerator(open(bundle_file.file_path, 'rb')))
                tmp_input_file = NamedTemporaryFile()
                source_map_processed_input_files.append(tmp_input_file)
                for chunk in output_pipeline:
                    m.update(chunk)
                    tmp_input_file.write(chunk)
                tmp_input_file.seek(0)
                infile_list.append(tmp_input_file.name)
            else:
                for chunk in FileChunkGenerator(open(bundle_file.file_path, 'rb')):
                    m.update(chunk)
                infile_list.append(bundle_file.file_path)

        hash_version = fixed_version or m.hexdigest()

        output_file_name = bundle.get_path(hash_version)
        if bundle.source_map_file_root and bundle.source_map_url_root:
            source_map_file_name = '%s.%s.%s.map' % (os.path.join(bundle.source_map_file_root, bundle.bundle_filename), hash_version, bundle.bundle_type)
            source_map_url = '%s.%s.%s.map' % (os.path.join(bundle.source_map_url_root, bundle.bundle_filename), hash_version, bundle.bundle_type)
        else:
            source_map_file_name = output_file_name + '.map'
            source_map_url = bundle.get_url(version=hash_version) + '.map'


        source_map_options = [
            '--source-map %s' % source_map_file_name,
            '--source-map-root %s' % bundle.source_map_files_url_root,
            '--source-map-url %s' % source_map_url,
            '-p %s' % os.path.realpath(bundle.files_root).count('/'),
            '-o %s' % output_file_name,
        ]

        # Consume the iterator into a zero length deque
        collections.deque(run_process(bundle.uglify_command.format(infile_list=' '.join(infile_list), source_map_options=' '.join(source_map_options))), maxlen=0)
    finally:
        for tmp_input_file in source_map_processed_input_files:
            tmp_input_file.close()

    return hash_version


def make_bundle(bundle, fixed_version=None):
    """
    Does all of the processing required to create a bundle and write it to disk, returning its hash version
    """
    tmp_output_file_name = '%s.%s.%s' % (os.path.join(bundle.bundle_file_root, bundle.bundle_filename), 'temp', bundle.bundle_type)

    iter_input = iter_bundle_files(bundle)

    output_pipeline = processor_pipeline(bundle.processors, iter_input)

    m = md5()

    with open(tmp_output_file_name, 'wb') as output_file:
        for chunk in output_pipeline:
            m.update(chunk)
            output_file.write(chunk)

    hash_version = fixed_version or m.hexdigest()

    output_file_name = bundle.get_path(hash_version)

    os.rename(tmp_output_file_name, output_file_name)

    return hash_version


class Command(BaseCommand):
    help = "Bundles up the media"
    requires_model_validation = False
    option_list = BaseCommand.option_list + (
        make_option('--dev',
            action='store_true',
            default=False,
            help='Do not version files' # This means repeated calls to bundle will just overwrite, and remove_bundles will remove the right things
        ),
    )

    def handle(self, *args, **options):
        self.stdout.write("Bundling...\n")
        dev_mode = bool(options.get('dev'))

        _bundle_versions = {}
        set_bundle_versions(_bundle_versions)

        for bundle in get_bundles():
            self.stdout.write("Writing bundle: %s\n" % bundle.name)

            if bundle.uglify_command:
                hash_version = make_uglify_bundle(bundle, fixed_version='_' if dev_mode else None)
            else:
                hash_version = make_bundle(bundle, fixed_version='_' if dev_mode else None)

            # Build bundle versions as we're going along in case they're used in templated bundles
            _bundle_versions[bundle.name] = hash_version

            self.stdout.write("\t%s\n" % bundle.get_version())

        version_info = '\n'.join(['    "%s": "%s",' % version for version in _bundle_versions.iteritems()])

        with open(bundles_settings.BUNDLES_VERSION_FILE, 'wb') as bundles_versions:
            bundles_versions.write("""\
#!/usr/bin/env python

BUNDLES_VERSIONS = {
%s
}
""" % version_info)

        for single_file_input, single_file_output in bundles_settings.BUNDLES_SINGLE_FILES:
            self.stdout.write("Writing: %s\n" % single_file_output)
            file_type = os.path.splitext(single_file_input)[1][1:]
            processors = processor_library.get_default_preprocessors_for(file_type) + processor_library.get_default_postprocessors_for(file_type)

            with open(single_file_output, 'wb') as output_file:
                for chunk in processor_pipeline(processors, FileChunkGenerator(open(single_file_input, 'rb'))):
                    output_file.write(chunk)

        self.stdout.write("Done.\n")



from subprocess import CalledProcessError
from django.core.management.base import BaseCommand, CommandError

from django_bundles.core import get_bundles
from django_bundles.processors import processor_pipeline
from django_bundles.utils.files import FileChunkGenerator
from django_bundles.utils.processes import run_process
from django_bundles.conf.bundles_settings import bundles_settings

from optparse import make_option
import collections

import os
from tempfile import NamedTemporaryFile


class Command(BaseCommand):
    help = "Lints the bundles based on settings.BUNDLES_LINTING"
    option_list = BaseCommand.option_list + (
        make_option('--failures-only',
            action='store_true',
            default=False,
            help='Only report failures'
        ),
        make_option('--pattern',
            help='Simple pattern matching for files',
        ),
    )
    requires_model_validation = False

    def lint_file(self, bundle_type, file_path, iter_input=None):
        command = bundles_settings.BUNDLES_LINTING[bundle_type]['command']

        input_file = None
        stdin = None

        if '{infile}' in command:
            if iter_input:
                if hasattr(iter_input, 'file_path'):
                    filename = iter_input.file_path
                else:
                    input_file = NamedTemporaryFile()
                    for chunk in iter_input:
                        input_file.write(chunk)
                    input_file.flush()

                    filename = input_file.name
            else:
                filename = file_path

            command = command.format(infile=filename)
        else:
            if iter_input:
                stdin = iter_input
            else:
                stdin = input_file = open(file_path, 'rb')

        try:
            # Consume the iterator into a zero length deque
            collections.deque(run_process(command, stdin=stdin, to_close=input_file), maxlen=0)
        except CalledProcessError as e:
            return False, e.output

        return True, ''


    def handle(self, *args, **options):
        self.show_successes = not bool(options.get('failures_only'))
        watch = bool(options.get('watch'))
        file_pattern = options.get('pattern')

        files_linted = set()

        failures = 0

        def file_checked(success, error_message, file_path):
            files_linted.add(file_path)

            if success:
                if self.show_successes:
                    self.stdout.write(self.style.HTTP_SUCCESS('OK\t\t%s\n' % file_path))
                return 0
            else:
                self.stdout.write(self.style.HTTP_SERVER_ERROR('FAIL\t\t%s\n' % file_path))
                self.stdout.write(self.style.HTTP_SERVER_ERROR(error_message))
                return 1


        for bundle in get_bundles():
            for bundle_file in bundle.files:
                if file_pattern and file_pattern not in bundle_file.file_path:
                    continue
                if bundle_file.file_path in files_linted:
                    continue

                # Check the file exists, even for non-linted files
                if not os.path.exists(bundle_file.file_path):
                    self.stdout.write(self.style.HTTP_SERVER_ERROR('FAIL\t\t%s\n' % bundle_file.file_path))
                    self.stdout.write(self.style.HTTP_SERVER_ERROR('File does not exist (referenced from %s)\n' % bundle.name))
                    failures += 1
                    continue

                if not bundle_file.lint:
                    continue

                success, error_message = self.lint_file(bundle.bundle_type, bundle_file.file_path, iter_input=processor_pipeline(bundle_file.processors, FileChunkGenerator(open(bundle_file.file_path, 'rb'))))

                failures += file_checked(success, error_message, bundle_file.file_path)


        for single_file_path, _ in bundles_settings.BUNDLES_SINGLE_FILES:
            success, error_message = self.lint_file(os.path.splitext(single_file_path)[1][1:], single_file_path)
            failures += file_checked(success, error_message, single_file_path)


        if failures:
            raise CommandError('%s FILE%s FAILED' % (failures, 'S' if failures > 1 else ''))
        else:
            self.stdout.write(self.style.HTTP_REDIRECT('\nALL FILES PASSED\n'))



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from django_bundles.conf.bundles_settings import bundles_settings
from django_bundles.utils.files import expand_file_names
from django_bundles.processors import processor_library

import os


class Bundle(object):
    """
    Representation of a bundle of files - configuration looks like:

    ('master_css', {                                                    # bundle name
        'type': 'css',                                                  # bundle type (e.g. css, js) - also the bundle file extension
        'uglify_command': None,                                         # special case to create a source map from uglify (this bypasses other post processors)
        'source_map_file_root': None,
        'source_map_url_root': None,
        'source_map_files_url_root': None,
        'files': (                                                      # list of files to include
            'css/*.css',                                                # pattern matching is done
            ('css/more/test3.css', {
                # bundle file options (see BundleFile class)
            }),
            'less/test.less',
        ),
        'files_url_root': settings.MEDIA_URL,                           # Root URL for the files in the bundle [OPTIONAL - defaults to settings.MEDIA_URL]
        'files_root': settings.MEDIA_ROOT,                              # Root path for the files in the bundle [OPTIONAL - defaults to settings.MEDIA_ROOT]
        'media': None,                                                  # Media type (e.g. screen, print) [OPTIONAL - defaults to None]
        'bundle_url_root': settings.MEDIA_URL,                          # Root URL for the bundle file [OPTIONAL - defaults to files_url_root]
        'bundle_file_root': settings.MEDIA_ROOT,                        # Root path for the bundle file [OPTIONAL - defaults to files_root]
        'bundle_filename': 'master_css',                                # Filename for the bundle (without extension) [OPTIONAL - defaults to the bundle name]
        'bundle_url': None,                                             # A full fixed URL for the bundle (overrides url_root and hashed file name)
        'bundle_file_path': None,                                       # To be used together with bundle_url to override the file root and hashed file name
        'processors': (                                                 # A list of post processors for the bundle (e.g. minifying) [OPTIONAL - defaults to the default processors for the bundle type]
            'django_bundles.processors.django_template.DjangoTemplateProcessor',                             # String,
            DjangoTemplateProcessor,                                                                         # Class,
            DjangoTemplateProcessor(),                                                                       # Instance,
            (ExecutableProcessor, {'command':'lessc {infile} {outfile}'}),                               # Class and kwargs
            ('django_bundles.processors.ExecutableProcessor', {'command':'lessc {infile} {outfile}'}),   # String and kwargs
        ),
        'precompile_in_debug': True,                                    # use watch_bundles management command to precompile files and serve them
    }),

    """
    def __init__(self, conf):
        """
        Initialize a bundle and it's BundleFiles based on a conf dict
        """
        self.name = conf[0]

        conf_dict = conf[1]

        # Basic settings and defaults
        self.bundle_type = conf_dict['type']
        self.files_url_root = conf_dict.get('files_url_root', settings.MEDIA_URL)
        self.files_root = conf_dict.get('files_root', settings.MEDIA_ROOT)
        self.uglify_command = conf_dict.get('uglify_command')
        self.source_map_file_root = conf_dict.get('source_map_file_root')
        self.source_map_url_root = conf_dict.get('source_map_url_root')
        self.source_map_files_url_root = conf_dict.get('source_map_files_url_root') or self.files_url_root
        self.media = conf_dict.get('media')
        self.bundle_url_root = conf_dict.get('bundle_url_root') or self.files_url_root
        self.bundle_file_root = conf_dict.get('bundle_file_root') or self.files_root
        self.bundle_filename = conf_dict.get('bundle_filename') or self.name
        self.precompile_in_debug = False if bundles_settings.GLOBAL_PRECOMPILE_DISABLE else conf_dict.get('precompile_in_debug', False)
        self.fixed_bundle_url = conf_dict.get('bundle_url')
        self.fixed_bundle_file_path = conf_dict.get('bundle_file_path')

        # Build the list of BundleFiles
        self.files = []
        self._bundle_files = {}

        for fileconf in list(conf_dict['files']):
            path, extra = fileconf, None
            # Each file definition can be a string or tuple containing the path and the conf dict
            if isinstance(fileconf, (tuple, list)):
                path = fileconf[0]
                extra = fileconf[1]

            # Expand *s in filenames
            try:
                for filename in expand_file_names(path, self.files_root):
                    bundle_file = BundleFile(filename, self.files_root, self.files_url_root, self.media, self.bundle_type, self.precompile_in_debug, extra=extra)
                    self.files.append(bundle_file)
                    self._bundle_files[bundle_file.file_path] = bundle_file
            except OSError:
                raise ImproperlyConfigured("Bundle %s - could not find file(s): %s" % (self.name, path))

        # Get the processors or use the default list
        if 'processors' in conf_dict:
            self.processors = processor_library.get_processors(conf_dict['processors'])
        else:
            self.processors = processor_library.get_default_postprocessors_for(self.bundle_type)

    def __contains__(self, filename):
        return filename in self._bundle_files

    def __getitem__(self, item):
        return self._bundle_files[item]

    def get_version(self):
        """
        Returns the current hash for this bundle
        """
        return get_bundle_versions().get(self.name)

    def get_url(self, version=None):
        """
        Return the filename of the bundled bundle
        """
        if self.fixed_bundle_url:
            return self.fixed_bundle_url
        return '%s.%s.%s' % (os.path.join(self.bundle_url_root, self.bundle_filename), version or self.get_version(), self.bundle_type)

    def get_path(self, hash_version):
        if self.fixed_bundle_file_path:
            return self.fixed_bundle_file_path
        return '%s.%s.%s' % (os.path.join(self.bundle_file_root, self.bundle_filename), hash_version, self.bundle_type)

    def get_file_urls(self):
        """
        Return a list of file urls - will return a single item if settings.USE_BUNDLES is True
        """
        if self.use_bundle:
            return [self.get_url()]
        return [bundle_file.file_url for bundle_file in self.files]

    @property
    def use_bundle(self):
        return bundles_settings.USE_BUNDLES and self.name not in bundles_settings.DEVELOPMENT_BUNDLES


class BundleFile(object):
    """
    Representation of a file in a bundle - configuration looks like:

    {
        'processors': (                                                                         # List of preprocessors (run before the file is concatenated into the bundle) [OPTIONAL - defaults to the default preprocessors based on file type]
            'django_bundles.contrib.processors.django_template.DjangoTemplateProcessor',        # Same as post processors: String, Class or Instance
        ),
        'lint': True,                                                                           # Whether to lint the file [OPTIONAL - defaults to settings.BUNDLES_LINTING[...]['default']]
        'type': 'js',                                                                           # File type [OPTIONAL - defaults to file extension]
    }

    """
    def __init__(self, filename, files_root, url_root, media, bundle_type, precompile_in_debug, extra=None):
        # basic settings
        self.bundle_type = bundle_type
        self.file_path = os.path.join(files_root, filename)
        self.file_url = os.path.join(url_root, filename)
        self.precompile_in_debug = precompile_in_debug
        if self.precompile_in_debug:
            self.precompile_url = os.path.join(url_root, '%s.%s' % (os.path.splitext(filename)[0], self.bundle_type))
            self.precompile_path = '%s.%s' % (os.path.splitext(self.file_path)[0], self.bundle_type)
        self.media = media

        # file_type (or take from extension)
        if extra and 'type' in extra:
            self.file_type = extra['type']
        else:
            self.file_type = os.path.splitext(filename)[1][1:]

        # Lint setting and default
        if bundle_type in bundles_settings.BUNDLES_LINTING:
            self.lint = bundles_settings.BUNDLES_LINTING[bundle_type].get('default', False)
            if extra and 'lint' in extra:
                self.lint = extra['lint']
        else:
            self.lint = False

        # Preprocessors or get defaults
        if extra and 'processors' in extra:
            self.processors = processor_library.get_processors(extra['processors'])
        else:
            self.processors = processor_library.get_default_preprocessors_for(self.file_type)


class BundleManager(object):
    """
    Kind of like an ordered dictionary, but not as complicated
    """
    def __init__(self):
        self.keys = []
        self._items = {}

    def __setitem__(self, key, value):
        if key not in self._items:
            self.keys.append(key)
        self._items[key] = value

    def __getitem__(self, item):
        return self._items[item]

    def __iter__(self):
        for key in self.keys:
            yield self._items[key]

    def items(self):
        return ((key, self._items[key]) for key in self.keys)


_cached_bundles = None
def get_bundles():
    """
    Used to cache the bundle definitions rather than loading from config every time they're used
    """
    global _cached_bundles

    if not _cached_bundles:
        _cached_bundles = BundleManager()

        for bundle_conf in bundles_settings.BUNDLES:
            _cached_bundles[bundle_conf[0]] = Bundle(bundle_conf)

    return _cached_bundles


_cached_versions = None
def get_bundle_versions():
    """
    Used to cache the bundle versions rather than loading them from the bundle versions file every time they're used
    """
    global _cached_versions
    if not bundles_settings.BUNDLES_VERSION_FILE:
        _cached_versions = {}
    if _cached_versions is None:
        locs = {}
        try:
            execfile(bundles_settings.BUNDLES_VERSION_FILE, locs)
            _cached_versions = locs['BUNDLES_VERSIONS']
        except IOError:
            _cached_versions = {}
    return _cached_versions


def set_bundle_versions(bundles_versions):
    """
    Used to update the cached versions whilst building the bundle
    """
    global _cached_versions
    _cached_versions = bundles_versions




from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase
from django.test.utils import override_settings
from django.template import Template, Context


from django_bundles.core import BundleManager, Bundle


import os


TEST_FILES_PATH = os.path.join(os.path.dirname(__file__), 'files')


@override_settings(BUNDLES=(
    ('test_bundle', {
        'type': 'css',
        'files': (
            'test1.css',
            'test2.css',
        ),
        'media': 'screen',
        'files_root': TEST_FILES_PATH,
    })
),
BUNDLES_TAG_HTML={ 'css': '<link href="%(file_url)s" type="text/css" rel="stylesheet"%(attrs)s />', })
class RenderBundleTest(TestCase):
    def test_render_tag(self):
        self.assertRaises(ImproperlyConfigured, Template("""{% load django_bundles_tags %}{% render_bundle "does_not_exist" %}""").render, Context())


class GetBundlesTest(TestCase):
    @override_settings(BUNDLES=( ('test_bundle', { 'type': 'css', 'files': (), }), ))
    def test_get_bundles(self):
        context = Context()
        Template("""{% load django_bundles_tags %}{% get_bundles as test_bundles %}""").render(context)

        self.assertTrue(isinstance(context.get('test_bundles'), BundleManager))
        self.assertTrue(isinstance(context['test_bundles']['test_bundle'], Bundle))



from subprocess import CalledProcessError
from django.core.management.base import BaseCommand, CommandError

from django_bundles.core import get_bundles
from django_bundles.processors import processor_pipeline
from django_bundles.utils.files import FileChunkGenerator
from django_bundles.utils.processes import run_process
from django_bundles.conf.bundles_settings import bundles_settings

from optparse import make_option
import collections

import os
from tempfile import NamedTemporaryFile


LOG_LINES = 10


class Command(BaseCommand):
    help = "Watches the bundles to lint and preprocess files"
    requires_model_validation = False

    def lint_file(self, full_path, bundle):
        try:
            iter_input = processor_pipeline(bundle[full_path].processors, FileChunkGenerator(open(full_path, 'rb')))
            command = bundles_settings.BUNDLES_LINTING[bundle.bundle_type]['command']

            input_file = None
            stdin = None

            if '{infile}' in command:
                if hasattr(iter_input, 'file_path'):
                    filename = iter_input.file_path
                else:
                    input_file = NamedTemporaryFile()
                    for chunk in iter_input:
                        input_file.write(chunk)
                    input_file.flush()

                    filename = input_file.name

                command = command.format(infile=filename)
            else:
                stdin = iter_input

            # Consume the iterator into a zero length deque
            collections.deque(run_process(command, stdin=stdin, to_close=input_file), maxlen=0)
            self.log_watch_result(full_path, True)
        except CalledProcessError as e:
            self.log_watch_result(full_path, False, error_message=e.output)

    def precompile(self, full_path, bundle):
        if bundle.precompile_in_debug and bundle[full_path].processors:
            try:
                with open(bundle[full_path].precompile_path, 'wb') as output_file:
                    for chunk in processor_pipeline(bundle[full_path].processors, FileChunkGenerator(open(full_path, 'rb'))):
                        output_file.write(chunk)
                self.log_precompile_result(full_path, True)
            except CalledProcessError:
                self.log_precompile_result(full_path, False)

    def drawscreen(self):
        import curses

        self.stdout.write(curses.tigetstr('clear'))
        self.stdout.write("Watching files for changes...")

        if self.errored_files:
            self.stdout.write(self.style.ERROR('%s file%s to fix' % (len(self.errored_files), 's' if len(self.errored_files) > 1 else '')))
        else:
            self.stdout.write(self.style.HTTP_REDIRECT('No files to fix'))

        self.stdout.write("\n")

        self.stdout.write(curses.tigetstr('cud1') * 2)

        for filename, error_message in self.errored_files.iteritems():
            self.stdout.write('\t' + self.style.HTTP_SERVER_ERROR(filename) + '\n')
            self.stdout.write(self.style.HTTP_SERVER_ERROR('\n'.join(['\t\t' + error_line for error_line in error_message.split('\n')])) + '\n\n')

        self.stdout.write(curses.tigetstr('cud1') * 2)

        for log_line in reversed(self.log_lines):
            self.stdout.write('\t' + log_line + '\n')

    def log_watch_result(self, src, result, error_message=None):
        if result:
            if src in self.errored_files:
                del self.errored_files[src]
            self.log_lines.append(self.style.HTTP_SUCCESS('OK\t\t%s' % src))
        else:
            self.log_lines.append(self.style.HTTP_SERVER_ERROR('FAIL\t\t%s' % src))
            self.errored_files[src] = self.style.HTTP_SERVER_ERROR(error_message)

        self.log_lines = self.log_lines[-LOG_LINES:]
        self.drawscreen()

    def log_precompile_result(self, src, success):
        if success:
            self.log_lines.append(self.style.HTTP_SUCCESS('PRECOMPILED\t\t%s' % src))
        else:
            self.log_lines.append(self.style.HTTP_SERVER_ERROR('PRECOMPILE FAILED\t\t%s' % src))
        self.log_lines = self.log_lines[-LOG_LINES:]
        self.drawscreen()

    def handle(self, *args, **options):
        try:
            import time
            from watchdog.observers import Observer
            from watchdog.events import FileSystemEventHandler
            import curses
        except ImportError:
            raise CommandError('watchdog is required for this (pip install watchdog')

        self.errored_files = {}
        self.log_lines = []
        watching = {}

        def check_and_lint_file(src):
            """
            Called when a watched file changes
            """
            full_path = os.path.realpath(src)

            for watchdir in watching:
                if watchdir in src:
                    for bundle in watching[watchdir]:
                        if full_path in bundle:
                            self.precompile(full_path, bundle)
                            self.lint_file(full_path, bundle)
                            return  # file could appear in multiple bundles

        class FileEventHandler(FileSystemEventHandler):
            def on_created(self, event):
                if not event.is_directory:
                    check_and_lint_file(event.src_path)

            def on_modified(self, event):
                if not event.is_directory:
                    check_and_lint_file(event.src_path)

        event_handler = FileEventHandler()
        observer = Observer()
        curses.setupterm()
        self.drawscreen()

        # First time we start, precompile everything
        initial_run = set()

        for bundle in get_bundles():
            # Note: watchdog seems to only work with relative paths
            watch_path = os.path.relpath(bundle.files_root)
            if watch_path not in watching:
                watching[watch_path] = set()
                observer.schedule(event_handler, path=watch_path, recursive=True)
            watching[watch_path].add(bundle)

            for bundle_file in bundle.files:
                if bundle_file.file_path in initial_run:
                    continue
                initial_run.add(bundle_file.file_path)
                self.precompile(bundle_file.file_path, bundle)

        observer.start()
        try:
            while True:
                time.sleep(10)
        except KeyboardInterrupt:
            observer.stop()
        observer.join()



import os
import fnmatch


def expand_file_names(path, files_root):
    """
    Expands paths (e.g. css/*.css in files_root /actual/path/to/css/files/)
    """
    # For non-wildcards just return the path. This allows us to detect when
    # explicitly listed files are missing.
    if not any(wildcard in path for wildcard in '*?['):
        return [path]
    else:
        dir_path, filename = os.path.split(path)
        return [os.path.join(dir_path, f) for f in fnmatch.filter(os.listdir(os.path.join(files_root, dir_path)), filename)]


class FileChunkGenerator(object):
    def __init__(self, input_file, chunk_size=1024, close=True):
        self.input_file = input_file
        self.chunk_size = chunk_size
        self.close = close
        self.file_path = input_file.name

    def __iter__(self):
        return self

    def next(self):
        chunk = self.input_file.read(self.chunk_size)
        if chunk == '':
            if self.close:
                self.input_file.close()
            raise StopIteration
        return chunk



import os
import uuid

from optparse import make_option

from django.core.management.base import BaseCommand

from django_bundles.core import get_bundles
from django_bundles.processors import processor_pipeline
from django_bundles.utils.files import FileChunkGenerator


class Command(BaseCommand):
    args = "target_directory"
    help = "Writes out files containing the list of input files for each bundle"
    requires_model_validation = False
    option_list = BaseCommand.option_list + (
        make_option('--bundle-type',
            help='Limit output to specific bundle types',
        ),
    )

    def handle(self, target_directory, *args, **options):
        try:
            os.mkdir(target_directory)
        except OSError:
            pass

        for bundle in get_bundles():
            if options.get('bundle_type') and bundle.bundle_type != options.get('bundle_type'):
                continue
            manifest_filename = os.path.join(target_directory, bundle.name) + '.manifest'
            with open(manifest_filename, 'w') as manifest:
                for bundle_file in bundle.files:
                    if bundle_file.processors:
                        # The file has a preprocessor. This means in its current state it may not be a valid file
                        # and thus not suitable for inclusion in the manifest. Do any appropriate preprocessing and
                        # write out an appropriate version
                        output_pipeline = processor_pipeline(bundle_file.processors, FileChunkGenerator(open(bundle_file.file_path, 'rb')))
                        output_file_name = os.path.realpath(os.path.join(target_directory, '%s-%s.%s' % (str(uuid.uuid4())[-8:], os.path.split(bundle_file.file_path)[1], bundle.bundle_type)))
                        with open(output_file_name, 'wb') as output_file:
                            for chunk in output_pipeline:
                                output_file.write(chunk)
                        manifest.write(output_file_name + "\n")
                    else:
                        manifest.write(bundle_file.file_path + "\n")



from django.conf import settings

USE_BUNDLES = not settings.DEBUG

DEVELOPMENT_BUNDLES = ()

DEFAULT_PREPROCESSORS = {
    'less': [
        ('django_bundles.processors.ExecutableProcessor', {'command':'lessc %(infile)s %(outfile)s'}),
    ],
}

DEFAULT_POSTPROCESSORS = {
    'js': [
        ('django_bundles.processors.ExecutableProcessor', {'command':'uglifyjs -nc --unsafe -o %(outfile)s %(infile)s'}),
    ],
}

BUNDLES_LINTING = {}

BUNDLES_LINT_SUCCESS_OK = True

BUNDLES_SINGLE_FILES = ()

BUNDLES_TAG_HTML = {
    'js': '<script src="%(file_url)s"></script>',
    'css': '<link href="%(file_url)s" type="text/css" rel="stylesheet"%(attrs)s />',
    'less': '<link href="%(file_url)s" type="text/less" rel="stylesheet/less"%(attrs)s />',
}

GLOBAL_PRECOMPILE_DISABLE = False


from django.test import TestCase
from django.core.exceptions import ImproperlyConfigured


from django_bundles.conf.bundles_settings import SettingsHelper


class ConfTest(TestCase):
    def setUp(self):
        class UserSettings():
            TEST1 = 'TEST1'

        class DefaultSettings():
            TEST1 = 'DEFAULT_TEST1'
            TEST2 = 'DEFAULT_TEST2'

        self.settings = SettingsHelper(UserSettings(), DefaultSettings(), ['TEST2'])


    def test_setting(self):
        self.assertEqual(self.settings.TEST1, 'TEST1')


    def test_default(self):
        self.assertEqual(self.settings.TEST2, 'DEFAULT_TEST2')


    def get_mandatory_setting(self):
        _ = self.settings.TEST3


    def test_mandatory(self):
        self.assertRaises(ImproperlyConfigured, self.get_mandatory_setting)


from django.test import TestCase


from django_bundles.utils import get_class


class GetClassTest(TestCase):
    def test_getting_class(self):
        cls = get_class('django.test.TestCase')
        cls2 = get_class('thisisnotaclass')
        cls3 = get_class('thisisnotaclass.alsothis')

        self.assertEqual(cls, TestCase)
        self.assertEqual(cls2, None)
        self.assertEqual(cls3, None)




from django import template

from django.utils.safestring import mark_safe
from django.core.exceptions import ImproperlyConfigured

from django_bundles.conf.bundles_settings import bundles_settings
from django_bundles.core import get_bundles

register = template.Library()


def _render_file(file_type, file_url, attrs=None):
    attr_string = ''
    if attrs:
        attr_string = ''.join(' %s="%s"' % x for x in attrs.iteritems())

    return bundles_settings.BUNDLES_TAG_HTML[file_type] % {
        'file_url': file_url,
        'attrs': attr_string,
    }


def _render_bundle(bundle_name):
    """
    Renders the HTML for a bundle in place - one HTML tag or many depending on settings.USE_BUNDLES
    """
    try:
        bundle = get_bundles()[bundle_name]
    except KeyError:
        raise ImproperlyConfigured("Bundle '%s' is not defined" % bundle_name)

    if bundle.use_bundle:
        return _render_file(bundle.bundle_type, bundle.get_url(), attrs=({'media':bundle.media} if bundle.media else {}))

    # Render files individually
    bundle_files = []

    for bundle_file in bundle.files:
        if bundle_file.precompile_in_debug:
            bundle_files.append(_render_file(bundle_file.bundle_type, bundle_file.precompile_url, attrs=({'media':bundle_file.media} if bundle.media else {})))
        else:
            bundle_files.append(_render_file(bundle_file.file_type, bundle_file.file_url, attrs=({'media':bundle_file.media} if bundle.media else {})))

    return '\n'.join(bundle_files)


@register.simple_tag
def render_bundle(bundle_name):
    return mark_safe(_render_bundle(bundle_name))


@register.assignment_tag(name='get_bundles')
def do_get_bundles():
    """
    Assigns the bundle definitions to a context variable
    """
    return get_bundles()



from django_bundles.conf.bundles_settings import bundles_settings
from django_bundles.utils import get_class
from django_bundles.utils.files import FileChunkGenerator
from django_bundles.utils.processes import run_process
from django.core.exceptions import ImproperlyConfigured

from tempfile import NamedTemporaryFile
import collections


class Processor(object):
    def __init__(self, **kwargs):
        for key, value in kwargs.iteritems():
            setattr(self, key, value)

    def process(self, iter_input):
        raise NotImplementedError


class ExecutableProcessor(Processor):
    command = None

    def process(self, iter_input):
        input_file = output_file = None
        stdin = iter_input
        format_kwargs = {}

        # Create temporary files for input and output if required
        if '{infile}' in self.command:
            stdin = None

            if hasattr(iter_input, 'file_path'):
                format_kwargs['infile'] = iter_input.file_path
            else:
                input_file = NamedTemporaryFile()
                format_kwargs['infile'] = input_file.name

                for chunk in iter_input:
                    input_file.write(chunk)

                input_file.seek(0) # Not sure why this is needed

        if '{outfile}' in self.command:
            output_file = NamedTemporaryFile()
            format_kwargs['outfile'] = output_file.name

        command = self.command.format(**format_kwargs)

        g = run_process(command, stdin=stdin, to_close=input_file)

        if output_file:
            # Consume the iterator into a zero length deque
            collections.deque(g, maxlen=0)
            return FileChunkGenerator(output_file)
        else:
            return g


def processor_pipeline(processors, iter_input):
    pipeline = iter_input

    for processor in processors:
        if not processor:
            continue

        pipeline = processor.process(pipeline)

    return pipeline


class ProcessorLibrary(object):
    def get_processor(self, processor_defn):
        class_path, init_kwargs = None, {}
        if isinstance(processor_defn, basestring):
            class_path = processor_defn
        elif isinstance(processor_defn, (list, tuple)):
            if len(processor_defn) == 1:
                class_path = processor_defn[0]
            elif len(processor_defn) == 2:
                class_path, init_kwargs = processor_defn

        if class_path:
            processor_class = get_class(class_path)

            if processor_class:
                return processor_class(**init_kwargs)

        raise ImproperlyConfigured("Invalid processor: %s" % repr(processor_defn))

    def get_processors(self, processors):
        return [instance for instance in (self.get_processor(processor) for processor in processors) if instance]

    def get_default_preprocessors_for(self, file_type):
        if file_type in bundles_settings.DEFAULT_PREPROCESSORS:
            return self.get_processors(bundles_settings.DEFAULT_PREPROCESSORS[file_type])
        return []

    def get_default_postprocessors_for(self, bundle_type):
        if bundle_type in bundles_settings.DEFAULT_POSTPROCESSORS:
            return self.get_processors(bundles_settings.DEFAULT_POSTPROCESSORS[bundle_type])
        return []

processor_library = ProcessorLibrary()